export SpectralElement

Base.@kwdef struct SpectralElement
  N::Int64                # polynomial order
  Npts::Int64             # number of quadrature points = N + 1
  z::Vector{Float64}      # quadrature points
  w::Vector{Float64}      # quadrature weights
  bw::Vector{Float64}     # barycentric weights
  D::Matrix{Float64}      # derivative matrix
  S::Matrix{Float64}      # stiffness matrix
  invM::Matrix{Float64}   # inverse mass matrix
  MDM::Matrix{Float64}    # = (M^-1 D^T M) = M^-1 S^T
  Ml_lhs::Vector{Float64} # = M^-1 l(-1)
  Ml_rhs::Vector{Float64} # = M^-1 l(+1)
  V::Matrix{Float64}      # = P_(j-1)(z_i), Vandermonde matrix
  invV::Matrix{Float64}   # inv(V), inverse Vandermonde matrix
  quadr::Symbol           # :LGL, :LGL_mass_lumped, :GLGL (deprecated, use gridtype instead)
  gridtype::Symbol        # = quadr
  iproduct::Function      # iproduct(f1, f2), inner product functor

  function SpectralElement(N::Int, quadr::Symbol)
    Npts = N + 1

    gridtype = quadr

    if quadr === :LGL
      z, w        = LGL.rule(Npts)
      purge_zeros!(z) # ensures that the centered node for even N basis sits exactly at x = 0.
      bw          = barycentric_weights(z)
      D           = first_derivative_matrix(z)
      S           = LGL.stiffness_matrix(Npts)
      invM        = LGL.inverse_mass_matrix_exact(Npts)
      iproduct    = (f1,f2) -> LGL.inner_product(z, w, f1 .* f2)
    elseif quadr === :LGL_lumped
      z, w        = LGL.rule(Npts)
      purge_zeros!(z) # ensures that the centered node for even N basis sits exactly at x = 0.
      bw          = barycentric_weights(z)
      D           = first_derivative_matrix(z)
      S           = LGL.stiffness_matrix(Npts)
      invM        = LGL.inverse_mass_matrix_lumped(Npts)
      iproduct    = (f1,f2) -> LGL.inner_product(z, w, f1 .* f2)
    elseif quadr === :GLGL
      z, w, v     = GLGL.rule(Npts)
      purge_zeros!(z) # ensures that the centered node for even N basis sits exactly at x = 0.
      bw          = barycentric_weights(z)
      D           = first_derivative_matrix(z)
      D2          = second_derivative_matrix(z)
      M           = GLGL.mass_matrix(z, w, v, D)
      S           = GLGL.stiffness_matrix(z, w, v, D, D2)
      invM        = inv(M)
      iproduct    = (f1,f2) -> GLGL.inner_product(z, w, v, D, f1, f2)
    else
      error("Invalid grid type specified! Use :LGL_lumped, :LGL, :GLGL.")
    end

    MDM     = invM * transpose(S)
    purge_zeros!(MDM)
    V       = vandermonde_matrix(z) # TODO needs to be replaced with vandermonde_matrix(z, basis_fn)
    invV    = inv(V)

    l_lhs   = zeros(N+1)
    l_rhs   = zeros(N+1)
    l_lhs[1]    = 1.0
    l_rhs[end]  = 1.0
    Ml_lhs = invM * l_lhs
    Ml_rhs = invM * l_rhs

    return new(N, Npts, z, w, bw, D, S, invM, MDM, Ml_lhs, Ml_rhs,
               V, invV, quadr, gridtype, iproduct)
  end
end
