export TensorBasis, TensorBasis1d, TensorBasis2d, TensorBasis3d,
       n_points, n_points_boundary, n_points_per_boundary,
       range_volume, range_boundary, strides_boundary


abstract type AbstractBasis end


struct TensorBasis{N_Dim}
  basis::NTuple{N_Dim,SpectralElement}

  function TensorBasis{N}(bases::Vararg{SpectralElement,N}) where N
    @assert 1 <= N <= 3 "Only support 1D, 2D, 3D rectangular basis!"
    return new{N}(bases)
  end
end

const TensorBasis1d = TensorBasis{1}
const TensorBasis2d = TensorBasis{2}
const TensorBasis3d = TensorBasis{3}


TensorBasis(se::SpectralElement...) = TensorBasis{length(se)}(se...)

Base.getindex(b::TensorBasis, idx::Int) = b.basis[idx]
# avoid bounds check with enums
# can we generalize Cart1d,Cart2d,Cart3d based on dimensions for other corodinates?
# Or do we want to keep cartesian coords as the computational coords intrinsic for the tensor basis?
# Base.getindex(b::TensorBasis1d, idx::Cart1d.Axis) = @inbounds b.basis[idx]
# Base.getindex(b::TensorBasis2d, idx::Cart2d.Axis) = @inbounds b.basis[idx]
# Base.getindex(b::TensorBasis3d, idx::Cart3d.Axis) = @inbounds b.basis[idx]


Base.ndims(b::TensorBasis{N_Dim}) where N_Dim = N_Dim
Base.size(b::TensorBasis) = Tuple(getproperty.(b.basis, :Npts))
n_points(b::TensorBasis) = prod(size(b))
n_points_boundary(b::TensorBasis1d) = 2
ratio_points(b::TensorBasis) = n_points_boundary(b) / n_points(b)
function n_points_boundary(basis::TensorBasis2d)
  sz = size(basis)
  return sum(2 .* sz) - 4
end
function n_points_boundary(basis::TensorBasis{3})
  sz = size(basis)
  return sum(2 .* (sz[1]*sz[2], sz[1]*sz[3], sz[2]*sz[3])) - 16
end
n_points_per_boundary(b::TensorBasis1d) = (1,)
function n_points_per_boundary(b::TensorBasis2d)
  Np1, Np2 = size(b)
  return (Np2,Np1)
end
function n_points_per_boundary(b::TensorBasis3d)
  Np1, Np2, Np3 = size(b)
  return (Np2*Np3,Np1*Np3,Np1*Np2)
end


function range_volume(tb::TensorBasis, offset=1)
  return Base.range(offset, length=offset+n_points(tb)-1)
end
function range_boundary(tb::TensorBasis, offset=1)
  return Base.range(offset, length=offset+n_points_boundary(tb))
end


# function strides_boundary(tb::TensorBasis1d, offset=1)
#   Np1, = size(tb)
#   return range(offset,length=1), range(offset+Np1-1,length=1)
# end
# function strides_boundary(tb::TensorBasis2d, offset=1)
#   npts = n_points(tb)
#   Np1, Np2 = size(tb)
#   return #= xmin =# range(offset,            length=Np1, step=1),
#          #= xmax =# range(offset+npts-Np1-1, length=Np1, step=1),
#          #= ymin =# range(offset,            length=Np2, step=Np1),
#          #= ymax =# range(offset+Np1-1,      length=Np2, step=Np1)
# end
# function strides_boundary(tb::TensorBasis3d, offset=1)
#   npts = n_points(tb)
#   Np1, Np2, Np3 = size(tb)
#   NNb1, NNb2, NNb3 = n_points_per_boundary(tb)
#   return #= xmin =# range(offset,             length=NNb1, step=1),
#          #= xmax =# range(offset+npts-NNb1-1, length=NNb1, step=1),
#          #= ymin =# range(offset,             length=NNb2, step=Np1),
#          #= ymax =# range(offset+Np1-1,       length=NNb2, step=Np1),
#          #= zmin =# range(offset,             length=NNb3, step=1),
#          #= zmax =# range(offset+npts-NNb1-1, length=NNb3, step=1)
# end


function Base.show(io::IO, tb::TensorBasis)
  print("""
        $(typeof(tb))
          number of dimensions              : $(ndims(tb))
          number of points per dimension    : $(size(tb))
          number of points per boundary
          and dimension (one sided)         : $(n_points_per_boundary(tb))
          total number of points            : $(n_points(tb))
          total number of boundary points   : $(n_points_boundary(tb))
          ratio boundary points/points      : $(ratio_points(tb))
        """)
end
